#!/usr/bin/perl

### pdv2svg - convert a rail track map to an SVG image.
### Copyright (C) 2009  ja:Matsukaze
### 
### This program is free software; you can redistribute it and/or
### modify it under the terms of the GNU General Public License
### as published by the Free Software Foundation; either version 2
### or version 3.
### 
### This program is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.

use Getopt::Std;
use utf8;
use Encode;
use Encode::Guess qw/ shiftjis euc-jp 7bit-jis /;
require "pdvdata.pl";

#get options
%opts = ();

getopts ("ac:d:hi:o:s:t:vAC", \%opts);
if($opts{h}){
    print<<EOM;
-a         : include definitions of all symbols.
-c CREATOR : creator information
-d "PIC1,PIC2,..." : include definitions of symbols in the list.
-h         : print this message.
-i FILE    : input file (default:STDIN)
-o FILE    : output file (default:STDOUT)
-s SCALE   : pixels per icon (default:30)
-t TITLE   : title 
-v         : verbose mode - print messages on STDERR.
-A         : abbreviation mode
-C         : color output
EOM
exit;
}
if($opts{s}){
    $scale = $opts{s};
}else{
    $scale = 30;
}
if($opts{i}){
    open(INFH,$opts{i}) or die "Cannot open file $opts{i}";
}else{
    open(INFH,"-");
}
if($opts{o}){
    $file = sprintf("> %s",$opts{o});
    open(OUTFH,$file) or die "Cannot open file $opts{o}";
}else{
    open(OUTFH,">-");
}
%used_pic=();
if($opts{d}){
    foreach $pic (split(/\,/,$opts{d})) {
	$used_pic{$pic} = 1;
    }
}

# read
$width=0;
$height=0;

if($opts{A}){
    while ( $line = <INFH> ){
	$height += 1;
	
	@row = split(/\|/,$line);
	$w = @row;
	if($width < $w){
	    $width = $w;
	}
	if($opts{v}){
	    for($i=0;$i<$w;$i++){
		print STDERR "$row[$i]";
		if($i != $w-1){
		    print STDERR " - ";
		}else{
		    print STDERR "\n";
		}
	    }
	}
	push(@list, [ @row ]);
	push(@flag_list, 0);
    }
    close(INFH);
}else{
    $content = join '\n', <INFH>;
    close(INFH);

    my $guess = Encode::Guess::guess_encoding($content);
    unless(ref $guess){
	# Failed to detect charcter code -- assume ascii.
	$content=decode("ascii",$content);
	if($opts{v}){
	    print STDERR "Unable to detect character code -- assume ascii\n";
	}
    }else{
	$content = $guess->decode($content);
	if($opts{v}){
	    printf STDERR ("Character code is %s\n",$guess->name);
	}
    }
    
    while ( $content =~ /\{\{(plan de voies|鉄道配線図|PV|plan de voies partiel|鉄道配線図部分|PVP)\|(.*)\}\}/gi ){
	my $line = $2;
	my $flag = 0;
	if( $1 =~ /plan de voies partiel|鉄道配線図部分|PVP/i ){
	    $flag = 1;
	    if($opts{v}){
		printf STDERR "(partial) ";
	    }
	}
	$height += 1;
	
	@row = split(/\|/,$line);
	$w = @row;
	if($width < $w){
	    $width = $w;
	}
	if($opts{v}){
	    for($i=0;$i<$w;$i++){
		print STDERR "$row[$i]";
		if($i != $w-1){
		    print STDERR " - ";
		}else{
		    print STDERR "\n";
		}
	    }
	}
	push(@list, [ @row ]);
	push(@flag_list, $flag);
    }
}

if($opts{v}){
    print(STDERR"width=$width, height=$height\n");
}

# find icons
for($i=0;$i<$height;$i++){
    $line = sprintf("\n <!-- row %d -->",$i+1);
    push(@output,$line);
    if( $flag_list[$i] ){
	push(@output," <!-- This row is generated by \"partial\" template -->");
    }
    $row = $list[$i];
    $w=$#{$row}+1;
    for ($j=0;$j<$w;$j++){
	$icon=$list[$i][$j];
	$icon =~ s/\s//g;
	unless($icon){
	    next;
	}
	$x = $j;
	$y = $i;
	undef $ref;
	if( exists($pdvdata::abbr{$icon}) ){
	    $icon = $pdvdata::abbr{$icon};
	}
	if( exists($pdvdata::table{$icon}) ){
	    $ref = $pdvdata::table{$icon};
	}
	if( ! defined($ref) ){
	    $line = sprintf(" <use xlink:href=\"#undefined\" x=\"%g\" y=\"%g\" width=\"2\" height=\"2\" /> <!-- %s -->",$x,$y,$list[$i][$j]);
	    push(@output,$line);
	    $line = sprintf("Undefined icon %s at x=%g y=%g\n",
			    $list[$i][$j],$x,$y);
	    print STDERR $line;
	    $used_pic{"undefined"} = 1;
	    next;
	}
	for($k=0;$k<=$#$ref;$k++){
	    $pic = $ref->[$k][0];
	    $used_pic{$pic} = 1;
	    @transf = ();
	    if($ref->[$k][1]){
		push(@transf,sprintf("rotate(%d,%g,%g)",
				     $ref->[$k][1],$x+1,$y+1));
	    }
	    if($ref->[$k][2]){
		push(@transf,sprintf("matrix(-1,0,0,1,%g,0)",2*$x+2));
	    }
	    if($#transf < 0){
		$transform = "";
	    }else{
		$transform = sprintf(" transform=\"%s\"",join(" ",@transf));
	    }
	    $line = sprintf(" <use xlink:href=\"#%s\" x=\"%g\" y=\"%g\" width=\"2\" height=\"2\"%s />",$pic,$x,$y,$transform);
	    push(@output,$line);
	}
    }
}

# write

## header
print OUTFH <<EOM;
<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">
EOM

if($opts{c}){
    print OUTFH "<!-- created by $opts{c} -->\n";
}

printf OUTFH ("<svg xmlns=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" width=\"%d\" height=\"%d\" version=\"1.1\">\n",$width*$scale,$height*$scale);

if($opts{t}){
    print OUTFH "<title> $opts{t} </title>\n";
}
##definitions
if($opts{C}){
    print OUTFH <<EOM;
<defs>
 <style type="text/css">
  <![CDATA[
   .str0 {stroke:none}
   .fil0 {fill:none}
   .filw {fill:white;stroke:none}
   .pont {stroke:darkslategray;stroke-width:2;stroke-linecap:butt;stroke-linejoin:miter;fill:none}
   .quai {stroke:none;fill:lightskyblue}
   .bv {stroke:none;fill:navy}
   .route {stroke:none;fill:maroon}
   .routec {stroke:white;stroke-width:2;stroke-dasharray:5;stroke-dashoffset:2.5;stroke-linecap:butt;stroke-linejoin:miter;fill:none}
   .pn {stroke:black;stroke-width:2;stroke-linecap:butt;stroke-linejoin:miter;fill:none}
   .fleche {stroke:black;stroke-width:2;stroke-linecap:butt;stroke-linejoin:miter;fill:none}
   .flechef {stroke:none;fill:black}
   .default {stroke:black;stroke-width:2;stroke-linecap:butt;stroke-linejoin:miter;fill:black}
  ]]>
 </style>
 <linearGradient id="Grad" gradientUnit="objectBoundingBox" spreadMethod="pad" >
  <stop stop-color="white" offset="0" />
  <stop stop-color="maroon" offset="1" />
 </linearGradient>
EOM
}else{
    print OUTFH <<EOM;
<defs>
 <style type="text/css">
  <![CDATA[
   .str0 {stroke:none}
   .fil0 {fill:none}
   .filw {fill:white;stroke:none}
   .pont {stroke:black;stroke-width:2;stroke-linecap:butt;stroke-linejoin:miter;fill:none}
   .quai {stroke:none;fill:#b3b3b3}
   .bv {stroke:none;fill:#666666}
   .route {stroke:none;fill:black}
   .routec {stroke:white;stroke-width:2;stroke-dasharray:5;stroke-dashoffset:2.5;stroke-linecap:butt;stroke-linejoin:miter;fill:none}
   .pn {stroke:#363636;stroke-width:2;stroke-linecap:butt;stroke-linejoin:miter;fill:none}
   .fleche {stroke:black;stroke-width:2;stroke-linecap:butt;stroke-linejoin:miter;fill:none}
   .flechef {stroke:none;fill:black}
   .default {stroke:black;stroke-width:2;stroke-linecap:butt;stroke-linejoin:miter;fill:black}
  ]]>
 </style>
 <linearGradient id="Grad" gradientUnit="objectBoundingBox" spreadMethod="pad" >
  <stop stop-color="white" offset="0" />
  <stop stop-color="black" offset="1" />
 </linearGradient>
EOM
}

##pictgrams
while( ($pic, $def) = each (%pdvdata::symbol) ){
    if($opts{a} || $used_pic{$pic}){
	print OUTFH " <symbol id=\"$pic\" viewBox=\"-15 -15 60 60\">";
	unless($used_pic{$pic}){
	    print OUTFH " <!-- This symbol is not used. -->";
	}
	print OUTFH "\n";
	print OUTFH $def;
	print OUTFH " </symbol>\n";
    }
}
print OUTFH "</defs>\n";

##main part
print OUTFH <<EOM;

<g class="default" transform="scale($scale,$scale) translate(-0.5,-0.5)">
 <rect class="filw" x="0.5" y="0.5" width="$width" height="$height" />
EOM
print OUTFH join("\n",@output);
print OUTFH "\n\n</g>\n</svg>\n";

close(OUTFH);
